import numpy as np

from chromatic_tda.utils.singleton import singleton
from chromatic_tda.utils.timing import TimingUtils


@singleton
class GeometricalUtils:

    def intersect_lines(self, p, p_dir, q, q_dir):
        """Return intersection of two lines in 2D given by a point and
        a direction.
        Return None if the directions are parallel.
        """
        TimingUtils().start("GeometricalUtils - intersect lines")
        t,s = np.linalg.solve(np.transpose((p_dir, -q_dir)), (q-p)) # p + t*p_dir = q + s*q_dir
        TimingUtils().stop("GeometricalUtils - intersect lines")
        return p + t*p_dir

    def project_to_line(self, p, p_dir, q):
        """Return orthogonal projection of point q into 2D line given
        by a point p and a direction p_dir."""
        TimingUtils().start("GeometricalUtils - project to line")
        res = self.intersect_lines(p, p_dir, q, np.array([-p_dir[1],p_dir[0]]))
        TimingUtils().stop("GeometricalUtils - project to line")
        return res

    def bisector(self, p, q):
        """Return the bisector of points p,q as a point and a direction.
        The point is the midpoint between p and q."""
        TimingUtils().start("GeometricalUtils - bisector")
        dif = q-p
        res = (q+p)/2, np.array([-dif[1], dif[0]])
        TimingUtils().stop("GeometricalUtils - bisector")
        return res

    def circum_center(self, p, q, r):
        """Return circum center of the triangle pqr."""
        TimingUtils().start("GeometricalUtils - circum center")
        res = self.intersect_lines(*self.bisector(p,q), *self.bisector(q,r))
        TimingUtils().stop("GeometricalUtils - circum center")
        return res

    def sq_dist(self, p, q):
        """Return squared distance of two points."""
        TimingUtils().start("GeometricalUtils - sq dist")
        if len(p)!=len(q):
            raise ValueError("Both points need to be of the same dimension.")
        res = sum(i**2 for i in p-q)
        TimingUtils().stop("GeometricalUtils - sq dist")
        return res

    def sq_dist_max(self, p, *args):
        """Return the max squared distance of p and points in args"""
        TimingUtils().start("GeometricalUtils - sq dist max")
        res = max( (self.sq_dist(p, q) for q in args), default=0 )
        TimingUtils().stop("GeometricalUtils - sq dist max")
        return res



import numpy as np
from chromatic_tda.utils.singleton import singleton


@singleton
class FloatingPointUtils:

    def is_trivial_bar(self, bar):
        return np.isclose(bar[0], bar[1])

    def ensure_smaller_or_equal(self, value : float, *coboundary_values) -> float:
        """Return `value` if it's smaller than the minimum of all the other passed values or close to it.
        Otherwise return the minimum of the other passed values"""
        minimum_above = min(coboundary_values, default=float('inf'))
        if (np.isclose(value, minimum_above) and value != minimum_above) or value > minimum_above:
            return minimum_above
        else:
            return value

    def ensure_weights_monotonicity_and_equal_values(self,
                                                     weight_function : dict[tuple[int, ...], float],
                                                     co_boundary: dict[tuple[int, ...], set]) -> None:
        for simplex in sorted(weight_function, key=len, reverse=True):
            weight_function[simplex] = self.ensure_smaller_or_equal(
                weight_function[simplex],
                *(weight_function[co_face] for co_face in co_boundary[simplex])
            )


class FloatComparisonWrapper:
    def __init__(self, num):
        self.value = num

    def __lt__(self, other):
        return self.__ne__(other) and self.value < other.value

    def __gt__(self, other):
        return self.__ne__(other) and self.value > other.value

    def __eq__(self, other):
        return np.isclose(self.value, other.value)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __le__(self, other):
        return self.__eq__(other) or self.value <= other.value

    def __ge__(self, other):
        return self.__eq__(other) or self.value >= other.value




import numpy as np
from matplotlib import pyplot as plt

from chromatic_tda.entities.simplicial_complex import SimplicialComplex
from chromatic_tda.utils.singleton import singleton


def plot_persistence_diagram(bars, ax=None, **kwargs):
    """Plot a single persistence diagram. Bars can be given in several different formats:
        - list of bars, i.e., (birth, death) tuples
        - list of (dimension, (birth, death)) tuples ... prints each dimension with different marker/color
        - dictionary of the form { dimension : list_of_bars }
    A matplotlib axis can be given as the ax argument.
    Various options can be given via keyword arguments, many coming from matplotlib options.
    """
    bars_dict = PlottingUtils().process_input_to_dict_form(bars)
    bars_dict_fin, bars_dict_inf = PlottingUtils().split_finite_infinite_dictionary(bars_dict)

    if 'lim' in kwargs and kwargs['lim'] is not None:
        xlim, ylim = PlottingUtils().parse_plot_limit_argument(kwargs['lim'])
    else:  # is not in kwargs (or is None)
        xlim, ylim = PlottingUtils().find_plot_limits(bars_dict_fin, **kwargs)

    infinity_position = kwargs.get('infinity_position', ylim[1] * .98)

    if ax:
        plt.sca(ax)
    else:
        plt.figure(figsize=(kwargs.get('size', 5), kwargs.get('size', 5)),
                   facecolor=kwargs.get('facecolor', 'white'))
    if not np.isclose(*xlim):  # to avoid passing identical low and high xlim to pyplot
        plt.xlim(xlim)
    if not np.isclose(*ylim):  # to avoid passing identical low and high ylim to pyplot
        plt.ylim(ylim)
    plt.gca().set_aspect(kwargs.get('aspect', 'equal'))
    plt.xticks(fontsize=kwargs.get('ticks_fontsize', 9))
    plt.yticks(fontsize=kwargs.get('ticks_fontsize', 9))
    if kwargs.get('axes_labels', True):
        plt.xlabel('birth', fontsize=kwargs.get('label_fontsize', 11))
        plt.ylabel('death', fontsize=kwargs.get('label_fontsize', 11))

    color = kwargs.get('color', {0: 'tab:green', 1: 'blue', 2: 'black', 3: 'orange'})
    marker = kwargs.get('marker', {0: 'D', 1: 'o', 2: 's', 3: '*'})
    if type(color) == list:
        color = {d: color[i] for i, d in enumerate(sorted(bars_dict)) if len(color) > i}
    elif type(color) == str:
        color = {d: color for d in bars_dict}

    if type(marker) == list:
        marker = {d: marker[i] for i, d in enumerate(sorted(bars_dict)) if len(marker) > i}
    elif type(marker) == str:
        marker = {d: marker for d in bars_dict}

    plt.axvline(0, color='gray', linewidth=1, alpha=.5)
    plt.plot(xlim, xlim, color='black', linewidth=1, alpha=.5)
    if 'title' in kwargs:
        plt.title(kwargs['title'], fontsize=kwargs.get('title_fontsize', 17))
    if not kwargs.get('only_finite', False) and any(len(bars_inf) > 0 for bars_inf in bars_dict_inf.values()):  # if also plotting infinite
        label = 'inf' if kwargs.get('label_inf', True) else None
        plt.axhline(infinity_position, color='tab:red', alpha=.5, linewidth=1, label=label)
        if label is not None:
            kwargs['infinite_line_legend'] = True
    for d in sorted(bars_dict_fin):
        if kwargs.get('only_finite', False):
            bars = bars_dict_fin[d]
        else:
            bars = list(bars_dict_fin[d]) + [[birth, infinity_position] for birth, death in bars_dict_inf[d]]
        plt.plot(
            *zip(*bars),
            linestyle='none',
            markersize=kwargs.get('markersize', 6),
            marker=marker.get(d, 'o'),
            color=color.get(d, None),
            fillstyle=kwargs.get('fillstyle', 'none'),
            alpha=kwargs.get('alpha', .7),
            label=str(d) if d != '' else None
        )
    if len(set(bars_dict_fin.keys()) - {''}) > 0 or kwargs.get('infinite_line_legend', False):
        plt.legend(loc='center right', fontsize=kwargs.get('legend_fontsize', 11))

    return plt.gcf(), plt.gca()


def plot_six_pack(data, axs=None, **kwargs):
    """Plot a six-pack of persistence diagrams. Data can be given in several different formats:
        - SimplicialComplex object
        - Dictionary with keys 'kernel', 'relative', 'cokernel', 'sub_complex', 'image', 'complex', and values
          the bars for the given group. The bars can be given in the following formats:
            - list of bars, i.e., (birth, death) tuples
            - list of (dimension, (birth, death)) tuples ... prints each dimension with different marker/color
            - dictionary of the form { dimension : list_of_bars }
    A list of six matplotlib axes can be given as the axs argument.
    Various options can be given via keyword arguments, many coming from matplotlib options.
    """
    if isinstance(data, SimplicialComplex):
        data = data.bars_six_pack()
    if isinstance(data, dict):
        data = {group : PlottingUtils().process_input_to_dict_form(bars) for group, bars in data.items()}
    else:
        raise TypeError("The data to plot a six-pack need to be given either as SimplicialComplex or a dictionary.")
    if axs is None:
        fig, axs = plt.subplots(2, 3,
                                figsize=(3 * kwargs.get('size', 5), 2 * kwargs.get('size', 5)),
                                facecolor=kwargs.get('facecolor', 'white'))
    axs = axs.flatten()
    if 'lim' in kwargs and kwargs['lim'] is not None:
        xlim, ylim = PlottingUtils().parse_plot_limit_argument(kwargs['lim'])
    else:
        xlims, ylims = zip(*[PlottingUtils().find_plot_limits(bars_dict) for group, bars_dict in data.items()])
        xlim = (min(x[0] for x in xlims), max(x[1] for x in xlims))
        ylim = (min(y[0] for y in ylims), max(y[1] for y in ylims))
    kwargs['lim'] = (xlim, ylim)

    if kwargs.get('axes_labels', True):
        plt.gcf().supxlabel('birth', fontsize=kwargs.get('label_fontsize', 14), y=.06)
        plt.gcf().supylabel('death', fontsize=kwargs.get('label_fontsize', 14), x=.09)
        kwargs['axes_labels'] = False

    groups_order = ['kernel', 'relative', 'cokernel', 'sub_complex', 'image', 'complex']
    if kwargs.get('allow_arbitrary_keys', False):
        groups = sorted(data.keys(), key=lambda k: groups_order.index(k) if k in groups_order else len(groups_order))
    else:
        groups = groups_order
        if set(groups) != set(data.keys()):
            raise ValueError("Keys in the data dictionary need to be 'kernel', 'relative', 'cokernel', 'sub_complex', "
                             "'image', 'complex'. To allow different keys, pass allow_arbitrary_keys=True.")

    for group, ax in zip(groups, axs):
        plot_persistence_diagram(data[group], ax=ax, title=group.replace('_', '-'), **kwargs)

    return plt.gcf(), axs



@singleton
class PlottingUtils:

    X_AXIS_EXTRA_RELATIVE_SPACE = 0.03
    Y_AXIS_EXTRA_RELATIVE_SPACE = 0.02
    Y_AXIS_EXTRA_RELATIVE_SPACE_WITH_INFINITY = 0.07

    def __init__(self):
        pass

    def find_max_finite_death_single_diagram(self, bars):
        return max((death for birth, death in bars if death < float('inf')), default=0)

    def find_max_finite_death_dictionary(self, bars_dict):
        return max((self.find_max_finite_death_single_diagram(bars) for bars in bars_dict.values()), default=0)

    def find_plot_limits(self, bars_dict, **kwargs):
        maxdeath = PlottingUtils().find_max_finite_death_dictionary(bars_dict)
        lim_left = -self.X_AXIS_EXTRA_RELATIVE_SPACE * maxdeath
        xlim = (lim_left, maxdeath * (1 + self.Y_AXIS_EXTRA_RELATIVE_SPACE))
        if kwargs.get('only_finite', False):
            ylim = (lim_left, maxdeath * (1 + self.Y_AXIS_EXTRA_RELATIVE_SPACE))
        else:
            ylim = (lim_left, maxdeath * (1 + self.Y_AXIS_EXTRA_RELATIVE_SPACE_WITH_INFINITY))

        return xlim, ylim

    def parse_plot_limit_argument(self, lim) -> tuple[tuple[float, float], tuple[float, float]]:
        if isinstance(lim, list) or isinstance(lim, tuple):
            if isinstance(lim[0], list) or isinstance(lim[0], tuple):
                xlim = lim[0]
                ylim = lim[1]
            else:  # is (xmax, ymax) tuple
                lim_left = -self.X_AXIS_EXTRA_RELATIVE_SPACE * max(lim[0], lim[1])
                xlim = (lim_left, lim[0])
                ylim = (lim_left, lim[1])
        else:  # is just one number
            lim_left = -self.X_AXIS_EXTRA_RELATIVE_SPACE * lim
            xlim = (lim_left, lim)
            ylim = (lim_left, lim)

        return xlim, ylim

    def convert_tuple_form_to_dict_form(self, bars_tuple_form):
        bars_dict_form = {}
        for dim, bar in bars_tuple_form:
            if dim not in bars_dict_form:
                bars_dict_form[dim] = []
            bars_dict_form[dim].append(bar)
        return bars_dict_form

    def process_input_to_dict_form(self, input_form):
        if not input_form:
            return {}
        if isinstance(input_form, dict):
            return input_form
        list_form = list(input_form)
        if len(list_form[0]) == 2:
            if isinstance(list_form[0][0], int):
                try:
                    second_coordinate_length = len(list_form[0][1])
                except TypeError:
                    second_coordinate_length = -1
                if second_coordinate_length == 2:
                    return self.convert_tuple_form_to_dict_form(list_form)
            return {'' : list_form}
        raise TypeError("Bars given in an invalid format. Should be either a dim -> list_of_bars dictionary"
                        " or a (dim, (birth, death)) list.")

    def split_finite_infinite_single_diagram(self, bars_list):
        return ([bar for bar in bars_list if bar[1] < float('inf')],
                [bar for bar in bars_list if bar[1] == float('inf')])

    def split_finite_infinite_dictionary(self, bars_dict):
        bars_finite = {}
        bars_infinite = {}
        for dim in bars_dict:
            bars_finite[dim], bars_infinite[dim] = self.split_finite_infinite_single_diagram(bars_dict[dim])
        return bars_finite, bars_infinite



import unittest

from chromatic_tda.entities.simplicial_complex import SimplicialComplex

class ComplexTest(unittest.TestCase):
    def test_complex(self, verbose=False) -> None:
        print("===== Testing Simplicial Complex has been started =====")

        complex = SimplicialComplex({(1,2,3),(2,3,4),(3,4,5),(4,6),(5,6),(3,7)})
        if verbose:
            complex.write()

        print(55*"=" + "\n")



from abc import ABC

class TopologicalFunctions(ABC):
    pass

class TopologicalFunctions2D(TopologicalFunctions):
    pass

class TopologicalFunctions3D(TopologicalFunctions):
    pass

class TopologicalFunctionFactory():
    def getInstance(self, dimension: str) -> TopologicalFunctions:
        if dimension is "2D":
            return TopologicalFunctions2D()
        if dimension is "3D":
            return TopologicalFunctions3D()
        
        raise NotImplementedError




from chromatic_tda.utils.topological_functions import TopologicalFunctions, TopologicalFunctionFactory

class BoundaryMatrix():
    def __init__(self):
        self.dimension = "2D"
        self.function_controller : TopologicalFunctions = TopologicalFunctionFactory().getInstance(self.dimension)



from time import perf_counter

from chromatic_tda.utils.singleton import singleton


@singleton
class TimingUtils:
    total_time_dict : dict[str, float]
    started_at_dict : dict[str, float]
    call_count_dict : dict[str, int]

    def __init__(self):
        self.flush()

    def flush(self):
        self.total_time_dict = {}
        self.started_at_dict = {}
        self.call_count_dict = {}

    def start(self, topic: str):
        self.started_at_dict[topic] = perf_counter()
        if topic in self.call_count_dict:
            self.call_count_dict[topic] += 1
        else:
            self.call_count_dict[topic] = 1

    def stop(self, topic: str):
        if topic in self.total_time_dict:
            self.total_time_dict[topic] += perf_counter() - self.started_at_dict[topic]
        else:
            self.total_time_dict[topic] = perf_counter() - self.started_at_dict[topic]

    def print(self):
        for topic in self.total_time_dict:
            print(f"Time [ {topic:<40} ] = {round(self.total_time_dict[topic], 2):<5} seconds (# calls : {self.call_count_dict[topic]:<6}, average = {round(self.total_time_dict[topic]/self.call_count_dict[topic], 3):<6})")



from chromatic_tda.utils.singleton import singleton

@singleton
class ReduceMatrixAlgorithm():

    ######################
    def reduce(self, matrix, order_function, order_function_row = None, return_V = False):
        '''
        Reduce given matrix. Using `order_function` to order columns.
        If `order_function_row` is given, it is used to order rows,
        otherwise `order_function` is also used for rows.
        The matrix is copied, i.e., given `matrix` remains unchanged.
        '''
        if order_function_row == None: order_function_row = order_function
        R = { k : v for k,v in matrix.items()}
        if return_V:
            V = { k : {k} for k in matrix}
        low_inv = {} # low_inv[i]=index of column with the lowest 1 at i
        for s in sorted(R, key=order_function):
            t = low_inv.get(max(R[s], key=order_function_row),-1) if len(R[s])!=0 else -1
            while t!=-1:
                R[s] = R[t]^R[s] # symmetric difference of t-th and s-th columns
                if return_V:
                    V[s] = V[t]^V[s]
                t = low_inv.get(max(R[s], key=order_function_row),-1) if len(R[s])!=0 else -1
            if len(R[s])!=0:
                low_inv[max(R[s], key=order_function_row)] = s

        return_dictionary = {}
        return_dictionary['R'] = R
        if return_V:
            return_dictionary['V'] = V
        return_dictionary['low_inv'] = low_inv
        return return_dictionary



import numpy as np
import itertools

from chromatic_tda.core.core_simplicial_complex import CoreSimplicialComplex
from chromatic_tda.utils.geometrical_utils import GeometricalUtils
from chromatic_tda.core.simplicial_complex_factory import CoreSimplicialComplexFactory


class CoreChromaticAlphaComplex():

    points : np.ndarray
    input_labels_to_internal_labels_dict : dict
    internal_labels_to_input_labels_dict : dict
    labels_number : int
    internal_labels : list
    simplicial_complex : CoreSimplicialComplex
    sq_rad : dict

    def __init__(self) -> None:
        self.input_labels_to_internal_labels_dict = {}
        self.labels_number = 0
        self.internal_labels = []
        self.sq_rad = {}

    def __iter__(self):
        yield from self.simplicial_complex

    def __len__(self) -> int:
        return len(self.simplicial_complex)

    def __contains__(self, element) -> bool:
        return element in self.simplicial_complex

    def _process_sub_complex_parameter_string(self, parameter : str):
        parameter = parameter.lower().strip()
        if parameter == 'all':
            list_of_sets_form = [set(self.input_labels_to_internal_labels_dict.keys())]
        elif parameter.endswith('chromatic'):
            chromaticity = parameter.replace('chromatic', '').replace('-', '')
            words_to_numbers = {'mono': 1, 'one': 1, 'bi': 2, 'two': 2, 'tri': 3, 'three': 3, 'tetra': 4, 'four': 4}
            if chromaticity in words_to_numbers:
                chromaticity = words_to_numbers[chromaticity]
            elif chromaticity.isnumeric() and chromaticity != '0':
                chromaticity = int(chromaticity)
            else:
                raise ValueError(f"The simplex restriction color parameter given `{parameter}` is invalid")
            list_of_sets_form = [set(x) for x in itertools.combinations(self.input_labels_to_internal_labels_dict.keys(), chromaticity)]
        else:
            list_of_sets_form = [{int(ch) for ch in w} for w in parameter.split(',')]

        return list_of_sets_form

    def _process_sub_complex_parameter(self, parameter, allow_unused_labels):
        if isinstance(parameter, str):
            list_of_sets_form = self._process_sub_complex_parameter_string(parameter)
        else:
            list_of_sets_form = [set(color_set) for color_set in parameter]

        if allow_unused_labels:
            list_of_sets_form = [color_set & set(self.input_labels_to_internal_labels_dict.keys()) for color_set in list_of_sets_form]
        else:
            for lab in set().union(*list_of_sets_form):
                if lab not in self.input_labels_to_internal_labels_dict.keys():
                    raise ValueError(f"There is no point labeled by `{lab}`. "
                                     f"To suppress this error, pass allow_unused_labels=True to get_complex function")
        list_of_sets_form_internal = [{self.input_labels_to_internal_labels_dict[lab] for lab in color_set}
                                      for color_set in list_of_sets_form]
        return list_of_sets_form_internal

    def simplex_labels_internal(self, simplex):
        return {self.internal_labels[v] for v in simplex}

    def simplex_labels_input(self, simplex):
        return {self.internal_labels_to_input_labels_dict[lab] for lab in self.simplex_labels_internal(simplex)}

    def get_complex(self, sub_complex, full_complex, relative, allow_unused_labels) -> CoreSimplicialComplex:
        if full_complex is None or full_complex == '':
            complex_simplices = set(self.simplicial_complex.boundary)
        else:
            complex_allowed_labels = self._process_sub_complex_parameter(full_complex, allow_unused_labels)
            complex_simplices = set()
            for simplex in self.simplicial_complex.boundary:
                simplex_labels = self.simplex_labels_internal(simplex)
                if any( simplex_labels.issubset(s) for s in complex_allowed_labels ):
                    complex_simplices.add(simplex)

        if relative is None or relative == '':
            relative_simplices = set()
        else:
            relative_labels_to_drop = self._process_sub_complex_parameter(relative, allow_unused_labels)
            relative_simplices = set()
            for simplex in self.simplicial_complex.boundary:
                simplex_labels = self.simplex_labels_internal(simplex)
                if any( simplex_labels.issubset(s) for s in relative_labels_to_drop ):
                    relative_simplices.add(simplex)

        restricted_complex : CoreSimplicialComplex = CoreSimplicialComplexFactory().create_restricted_instance(self.simplicial_complex, complex_simplices - relative_simplices)

        if sub_complex is None or sub_complex == '':
            restricted_complex.set_sub_complex(set())
        else:
            sub_complex_allowed_labels = self._process_sub_complex_parameter(sub_complex, allow_unused_labels)
            sub_complex_simplices = set()
            for simplex in restricted_complex.boundary: # already relativized in definition of new-complex
                simplex_labels = self.simplex_labels_internal(simplex)
                if any( simplex_labels.issubset(s) for s in sub_complex_allowed_labels ):
                    sub_complex_simplices.add(simplex)
            restricted_complex.set_sub_complex(sub_complex_simplices)

        return restricted_complex

    def star_vertices(self, simplex) -> set:
        return set().union(*self.simplicial_complex.co_boundary[simplex]) - set(simplex)

    def is_empty_stack(self, center, simplex) -> bool:
        """
        Return True if the stack centered at `center` is empty.
        Assume that the center given is a circum-center for each
        monochromatic face, and check whether the vertices in
        co-boundary of simplex are inside the circum-centers of
        the corresponding colors. Colors not present in simplex
        are treated as automatically satisfied.
        """
        vertices_to_check = self.star_vertices(simplex)
        for mono_chrom_face in self.split_simplex(simplex): # for every color
            if len(mono_chrom_face) > 0: # if the color is present in simplex
                radius = GeometricalUtils().sq_dist(center, self.points[mono_chrom_face[0]])
                if any( (self.internal_labels[mono_chrom_face[0]] == self.internal_labels[v] and
                         GeometricalUtils().sq_dist(center, self.points[v]) < radius)
                        for v in vertices_to_check ):
                    return False
        return True

    def split_simplex(self, simplex) -> list:
        return [tuple(i for i in simplex if self.internal_labels[i] == 0),
                tuple(i for i in simplex if self.internal_labels[i] == 1),
                tuple(i for i in simplex if self.internal_labels[i] == 2)]

    def split_simplex_sort_by_size(self, simplex) -> list:
        return sorted(self.split_simplex(simplex), key=len, reverse=True)

    def write(self) -> None:
        print()
        print(f"**** Delaunay Complex (dimension = {self.simplicial_complex.get_dimension()}) ****")
        for (p,l) in zip(self.points, self.internal_labels):
            print(f"{p} -> {l}")
        print(42*"*")
        print()



# from abc import ABC, abstractmethod
import numpy as np
import random
from scipy.spatial import Delaunay

from chromatic_tda.utils.singleton import singleton
from chromatic_tda.core.core_chromatic_alpha_complex import CoreChromaticAlphaComplex
from chromatic_tda.utils.boundary_matrix_utils import BoundaryMatrixUtils
from chromatic_tda.algorithms.radius_function_utils import RadiusFunctionUtils
from chromatic_tda.core.simplicial_complex_factory import CoreSimplicialComplexFactory
from chromatic_tda.utils.timing import TimingUtils



@singleton
class CoreChromaticAlphaComplexFactory():

    def create_instance(self, points, labels, lift_perturbation, point_perturbation, **kwargs) -> CoreChromaticAlphaComplex:
        alpha_complex = CoreChromaticAlphaComplex()
        self.build_alpha_complex(alpha_complex, points, labels, lift_perturbation, point_perturbation, **kwargs)
        return alpha_complex

    def build_alpha_complex(self, alpha_complex: CoreChromaticAlphaComplex, points, labels, lift_perturbation,
                            point_perturbation, **kwargs):
        """
        Compute the chromatic alpha complex of given points.
        At most three different labels allowed
        """
        TimingUtils().start("Build Alpha Complex")

        if point_perturbation:
            alpha_complex.points = np.array([[p + point_perturbation * (random.random() - .5) for p in pt] for pt in points])
        else:
            alpha_complex.points = np.array(points)
            
        if alpha_complex.points.shape[1] != 2:
            raise ValueError("Points has to be an iterable of two-dimensional points.")

        alpha_complex.input_labels_to_internal_labels_dict = {lab : i for i, lab in enumerate(sorted(set(labels)))}
        alpha_complex.internal_labels_to_input_labels_dict = {i : lab for lab, i in alpha_complex.input_labels_to_internal_labels_dict.items()}
        alpha_complex.labels_number = len(alpha_complex.input_labels_to_internal_labels_dict)
        alpha_complex.internal_labels = [alpha_complex.input_labels_to_internal_labels_dict[lab] for lab in labels]

        if alpha_complex.labels_number > 3:
            raise ValueError(f"There can be at most 3 different labels, {alpha_complex.labels_number} given.")
        if len(alpha_complex.points) != len(alpha_complex.internal_labels):
            raise ValueError("The list of labels must have the same length as the list of points.")

        self.construct_chromatic_delaunay(alpha_complex, lift_perturbation)
        alpha_complex.simplicial_complex.co_boundary = BoundaryMatrixUtils().make_co_boundary(alpha_complex.simplicial_complex.boundary)
        RadiusFunctionUtils().compute_radius_function(alpha_complex, **kwargs)
#        RadiusFunctionParallelUtils(alpha_complex).compute_radius_function_in_parallel(**kwargs)

        TimingUtils().stop("Build Alpha Complex")

    def construct_chromatic_delaunay(self, alpha_complex: CoreChromaticAlphaComplex, lift_perturbation = None):
        TimingUtils().start("Construct Chromatic Delaunay")

        del_complex = Delaunay( self.chromatic_lift(alpha_complex, lift_perturbation))
        all_labels = set(alpha_complex.internal_labels)
        simplices = [simplex for simplex in del_complex.simplices if set(alpha_complex.internal_labels[i] for i in simplex) == all_labels]
        alpha_complex.simplicial_complex = CoreSimplicialComplexFactory().create_instance(simplices)

        TimingUtils().stop("Construct Chromatic Delaunay")

    def chromatic_lift(self, alpha_complex: CoreChromaticAlphaComplex, lift_perturbation = None):
        """
        Add extra coordinates to lift points to the chromatic simplex. Here we choose one-hot embedding without
        the first coordinate. That is, 0 --> (0,0,0,...), 1 --> (1,0,0,...), 2 --> (0,1,0,...), etc.
        """
        pts_lift = np.array([
            np.concatenate((point, [1 if i == label else 0 for i in range(1, alpha_complex.labels_number)]))
            for point, label in zip(alpha_complex.points, alpha_complex.internal_labels)])
        if lift_perturbation:
            for i in range(len(pts_lift)):
                pts_lift[i] += [0, 0] + [lift_perturbation * random.random() for i in range(1, alpha_complex.labels_number)]

        return pts_lift



from chromatic_tda.utils.singleton import singleton
from chromatic_tda.utils.timing import TimingUtils

@singleton
class BoundaryMatrixUtils():
    def __len__(self, boundary):
        return len(boundary)

    def __contains__(self, boundary, item):
        return item in boundary
    
    def __iter__(self, boundary):
        return (simplex for simplex in boundary)

    def make_co_boundary(self, boundary):     # add_co_bnd function
        TimingUtils().start("Make Co-Boundary")

        co_boundary = { simplex : set() for simplex in boundary}
        for simplex in boundary:
            for face in boundary[simplex]:
                co_boundary[face].add(simplex)

        TimingUtils().stop("Make Co-Boundary")
        return co_boundary

    def get_restricted_boundary(self, complex, simplices_set: set):
        return { simplex : complex.boundary[simplex] & simplices_set for simplex in set(complex.boundary) & simplices_set }

    def print_matrix(self, matrix, order_function):
        order = sorted(matrix, key=order_function)
        for row in order:
            for column in order:
                print('1' if row in matrix[column] else '0', end='')
            print()
